##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `examples` package provides resources for learning and setting up Merlin workflows.

Modules:
    examples.py: Contains example specification files for Merlin workflows, along with
        detailed explanations of each block in the specification.
    generator.py: Provides utilities for managing and generating example workflows.
"""
