##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `server` package defines the functionality for managing a containerized server
through Merlin.

Modules:
    server_commands.py: Main functions to interact with the server.
    server_config.py: Server configuration functions.
    server_util.py: Defines the structure of our server configurations.
"""
